#include "globals.h"

typedef struct NETWORK_CELL
{
    int isSpace;
    TASK *task;

    struct NETWORK_CELL *prev;
    struct NETWORK_CELL *next;
}
NETWORK_CELL;


extern char *monthNames[];
extern char *dayNames[];

static NETWORK_CELL **network_cols;
static int numX;
static int numY;

static TASKLIST *startTL;
static TASKLIST *finishTL;

/* construct start and finish markers */
static void constructStartFinish()
{
    TASK *t;

    t = (TASK *)malloc(sizeof(TASK));
    if ( t == NULL )
	Error("Can't malloc start task in AddTask()");

    t->id = strdup("");
    t->name = strdup("START");
    t->depends = NULL;
    t->start = 0;
    t->nx = 0;
    t->ny = pc_start_ny;
    
    startTL = (TASKLIST *)malloc(sizeof(TASKLIST));
    if ( startTL == NULL )
	Error("Can't malloc startTL in constructStartFinish()");
    startTL->task = t;
    startTL->next = NULL;

    t = (TASK *)malloc(sizeof(TASK));
    if ( t == NULL )
	Error("Can't malloc finish task in AddTask()");

    t->id = strdup("");
    t->name = strdup("FINISH");
    t->depends = NULL;
    t->start = 0;
    t->nx = numX - 1;
    t->ny = pc_finish_ny;
    
    finishTL = (TASKLIST *)malloc(sizeof(TASKLIST));
    if ( finishTL == NULL )
	Error("Can't malloc finishTL in constructStartFinish()");
    finishTL->task = t;
    finishTL->next = NULL;
}

    
/* return 1 if nx not assigned yet but all dependencies have been assigned */
static int canAssignX(TASK *task)
{
    TASKLIST *dtl;

    if ( task->nx != -1)
	return 0;			/* already assigned */

    if ( strncasecmp(task->id,"vac_",4) == 0 )
	return 0;			/* vacations aren't assigned */

    for ( dtl = task->depends ; dtl != NULL ; dtl = dtl->next )
	if ( dtl->task->nx == -1 )
	    return 0;

    return 1;
}


/* set value of nx to be 1 more than child values */
static void assignX(TASK *task)
{
    TASKLIST *dtl;
    int max_depend;

    /* the "start" task will be in column 0 so all true tasks must be
       in column 1 or greater */
    max_depend = 0;
    for ( dtl = task->depends ; dtl != NULL ; dtl = dtl->next )
	if ( dtl->task->nx > max_depend )
	    max_depend = dtl->task->nx;

    task->nx = max_depend + 1;
}


/* find maximum X value */
static int maximumX(void)
{
    TASKLIST *tl;
    int maxX;

    maxX = -1;
    for ( tl = tasklist ; tl != NULL ; tl = tl->next )
    {
	if ( strncasecmp(tl->task->id,"vac_",4) == 0 )
	    continue;
	
	if ( tl->task->nx > maxX )
	    maxX = tl->task->nx;
    }
    if ( maxX == -1 )
	Error("Can't find maxX in maximuX()");

    return maxX;
}


static int nextAvailableY(NETWORK_CELL *clist, int y)
{
    NETWORK_CELL *pc;
    int ny;

    ny = y;
    while (1)
    {
	ny++;
	for ( pc = clist; pc != NULL; pc = pc->next )
	    if ( pc->task->ny == ny )
		break;
	if ( pc == NULL )
	    return ny;
    }
    return 0;				/* not reached */
}


/* assign Y values */
static void assignY(void)
{
    TASKLIST *tl;
    int y;
    int x;
    NETWORK_CELL *pc;
    NETWORK_CELL *cell;

    /* find maximum X values */
    numX = maximumX();

    // printf("maxX = %d\n", numX);

    numX++;				/* convert to number of x vals */

    numX++;				/* to allow for finish in last col */

    if ((network_cols = (NETWORK_CELL **)malloc(sizeof(NETWORK_CELL *) * numX)) == NULL)
	Error("assignY: Couldn't alloc network_cols");
    for ( x = 0; x < numX ; x++ )
	network_cols[x] = NULL;

    /* build up an index of cells by column */
    for ( tl = tasklist ; tl != NULL ; tl = tl->next )
    {
	if ( strncasecmp(tl->task->id,"vac_",4) == 0 )
	    continue;

	x = tl->task->nx;

	if ( ( cell = (NETWORK_CELL *)malloc(sizeof(NETWORK_CELL))) == NULL)
	    Error("assignY: Can't alloc network cell");
	cell->task = tl->task;
	cell->isSpace = 0;
	cell->prev = NULL;
	cell->next = NULL;

	if ( network_cols[x] == NULL )
	    network_cols[x] = cell;
	else
	{
	    for ( pc = network_cols[x]; pc->next != NULL; pc = pc->next )
		;
	    cell->prev = pc;
	    pc->next = cell;
	}
    }
    
    /* assign Y values on first come basis, skipping those assigned */
    for ( x = 0; x < numX; x++ )
    {
	y = nextAvailableY(network_cols[x], -1);
	for ( pc = network_cols[x]; pc != NULL; pc = pc->next )
	    if ( pc->task->ny == -1 )
	    {
		pc->task->ny = y;
		y = nextAvailableY(network_cols[x], y);
	    }
    }

    /* find number of Y values */
    numY = 0;
    for ( x = 0; x < numX; x++ )
	for ( pc = network_cols[x]; pc != NULL; pc = pc->next )
	    if ( pc->task->ny > numY )
		numY = pc->task->ny;
    numY++;
}


FILE *openNetworkDiagram(char *filename, int llx, int lly, int urx, int ury)
{ 
    FILE *f;

    if ( ( f = fopen(filename, "w") ) == NULL )
	Error("Cannot open file [%s]",filename);

    fprintf(f,"%%!\n");
    fprintf(f,"%%%%BoundingBox: %d %d %d %d\n",llx,lly,urx,ury);

    fprintf(f,"/OUTLINE\n");
    fprintf(f,"{\n");
    fprintf(f,"  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n");
    fprintf(f,"  x1 y1 moveto\n");
    fprintf(f,"  x2 y1 lineto\n");
    fprintf(f,"  x2 y2 lineto\n");
    fprintf(f,"  x1 y2 lineto\n");
    fprintf(f,"  closepath\n");
    fprintf(f,"  0 setgray\n");
    fprintf(f,"  stroke\n");
    fprintf(f,"} def\n");

    fprintf(f,"/BACKGROUND\n");
    fprintf(f,"{\n");
    fprintf(f,"  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n");
    fprintf(f,"  x1 y1 moveto\n");
    fprintf(f,"  x2 y1 lineto\n");
    fprintf(f,"  x2 y2 lineto\n");
    fprintf(f,"  x1 y2 lineto\n");
    fprintf(f,"  closepath\n");
    fprintf(f,"  %f setgray\n", tg_white);
    fprintf(f,"  fill\n");
    fprintf(f,"} def\n");

    fprintf(f,"/%s findfont %f scalefont setfont\n",
	    pc_fontname1, pc_fontsize1 );

    fprintf(f,"0 setgray\n");
    fprintf(f,"0.01 setlinewidth\n");

    return f;
}


static NETWORK_CELL *findCellInColumn(int x, char *id)
{
    NETWORK_CELL *cell;

    for ( cell = network_cols[x] ; cell != NULL ; cell = cell->next )
    {
	if ( cell->isSpace )
		continue;
	if ( strcmp( cell->task->id, id ) == 0 )
	    return cell;
    }

    Error("Asked for cell (%s) in col (%d) but not there", id, x );
    return NULL;			/* only to keep compiler happy */
}


static void write_milestones(FILE *fp, int ax, int bx, int ay, int by)
{
    MILESTONELIST *ml;
    MILESTONE *m;
    TASK *t;
    int px, py;
    char buf[1024];
    
    for ( ml = milestonelist; ml != NULL; ml = ml->next)
    {
	m = ml->milestone;
	t = m->critical;

	px = ax * t->nx + bx + pc_width;
	py = ay * t->ny + by + pc_height;

	fprintf(fp, "%d %d moveto ", px + pc_height / 8, py + pc_height / 4);
	fprintf(fp, "%d %d lineto ", px, py);
	fprintf(fp, "%d %d lineto ", px - pc_height / 8, py + pc_height / 4);
	fprintf(fp, "%d %d lineto ", px + pc_width / 3, py + pc_height / 4);
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px - pc_height / 8,
		py + pc_height / 4 + pc_textup );
	if (milestone_ids)
	    sprintf(buf, "%s %s", m->id, m->name);
	else
	    sprintf(buf, "%s", m->name);
	fprintf(fp, "(%s) show\n", buf);
    }
}

	

void write_chart(int start, int finish, char *filename)
{
    int minX;
    int maxX;
    int minY;
    int maxY;
    int sizeX;
    int sizeY;
    int x;
    NETWORK_CELL *cell;
    int px, py;				/* coords in postscript */
    int ax, bx, ay, by;
    int llx, lly, urx, ury;
    FILE *fp;
    char buf[1024];
    int pxd, pyd;
    TASKLIST *dtl;
    NETWORK_CELL *pc;
    int c0, c1, c2;
    int r0, r1, r2, r3;
    double dTextHeight;

    /* work out the size */
    if ( start > 0 )
	minX = start;
    else
	minX = 0;
    if ( finish < numX )
	maxX = finish;
    else
	maxX = numX - 1;
    sizeX = maxX - minX + 1;
    minY = numY;
    maxY = 0;
    for ( x = minX ; x <= maxX ; x++)
	for ( cell = network_cols[x] ; cell != NULL ; cell = cell->next )
	{
	    if ( cell->task->ny < minY )
		minY = cell->task->ny;
	    if ( cell->task->ny > maxY )
		maxY = cell->task->ny;
	}
    sizeY = maxY - minY + 1;

//    printf("(minX,maxX,sizeX)=(%d,%d,%d)\n", minX, maxX, sizeX);
//    printf("(minY,maxY,sizeY)=(%d,%d,%d)\n", minY, maxY, sizeY);


    /* work out map from (x,y) to postscript coords of bottom left corner */
    ax = pc_width + pc_space;
    bx = -ax * minX;
    ay = -( pc_height + pc_space );
    by = ( pc_height + pc_space ) * maxY;
    
    /* work out bounding box */
    llx = 0;
    lly = -1;				/* to allow for line width */
    urx = ax * maxX + bx + pc_width + 1; /* +1 to allow for line width */
    ury = ay * minY + by + pc_height;
    
    /* allow for potential milestone marks on top row */
    urx += pc_width / 3;
    ury += pc_height / 2;

    fp = openNetworkDiagram(filename, llx, lly, urx, ury);

    /* do all of the joins first */
    for ( x = minX ; x <= maxX ; x++)
    {
	px = ax * x + bx;
	for ( cell = network_cols[x] ; cell != NULL ; cell = cell->next )
	{
	    if ( cell->isSpace )
		continue;
	    
	    py = ay * cell->task->ny + by;

	    if ( cell->task->depends == NULL )
	    {
		if ( minX == 0 )
		{
		    pxd = ax * startTL->task->nx + bx;
		    pyd = ay * startTL->task->ny + by;
		    
		    fprintf(fp, "%d %d moveto\n", px, py + pc_height / 2 );
		    fprintf(fp, "%d %d lineto\n",
			    pxd + pc_width, pyd + pc_height / 2 );
		    fprintf(fp, "stroke\n" );
		}
	    }
	    else
	    {
		for ( dtl = cell->task->depends; dtl != NULL ; dtl = dtl->next )
		{
		    pc = findCellInColumn( dtl->task->nx , dtl->task->id );
		    pxd = ax * pc->task->nx + bx;
		    pyd = ay * pc->task->ny + by;
		    
		    fprintf(fp, "%d %d moveto\n", px, py + pc_height / 2 );
		    fprintf(fp, "%d %d lineto\n",
			    pxd + pc_width, pyd + pc_height / 2 );
		    fprintf(fp, "stroke\n" );
		}
	    }
	}
    }
    if ( maxX == numX - 1 )
    {
	px = ax * maxX + bx;
	py = ay * finishTL->task->ny + by;
	for ( x = minX ; x <= maxX ; x++)
	{
	    for ( cell = network_cols[x] ; cell != NULL ; cell = cell->next )
	    {
		if ( cell->isSpace )
		    continue;

		if ( cell->task->follows == NULL )
		{
		    pxd = ax * cell->task->nx + bx;
		    pyd = ay * cell->task->ny + by;
		    
		    fprintf(fp, "%d %d moveto\n", px, py + pc_height / 2 );
		    fprintf(fp, "%d %d lineto\n",
			    pxd + pc_width, pyd + pc_height / 2 );
		    fprintf(fp, "stroke\n" );
		}
	    }
	}
    }

    
    c0 = 0;
    c1 = pc_width / 3;
    c2 = 2 * pc_width / 3;
    dTextHeight = 0.25 * ( pc_height - 7 * pc_textup );
    r0 = 0;
    r1 = 2 * pc_textup + (int)dTextHeight;
    r2 = 3 * pc_textup + (int)( 2.0 * dTextHeight );
    r3 = 5 * pc_textup + (int)( 3.0 * dTextHeight );

    fprintf(fp,"1.0 setlinewidth\n");

    /* do the start marker */
    if ( minX == 0 )
    {
	x = 0;
	px = ax *  x + bx;
	py = ay * startTL->task->ny + by;
	fprintf(fp, "%d %d %d %d BACKGROUND\n",
		px, px + pc_width, py, py + pc_height);

	fprintf(fp, "%d %d %d %d OUTLINE\n",
		px, px + pc_width, py, py + pc_height);

	fprintf(fp, "%d %d moveto ", px, py + r1 );
	fprintf(fp, "%d %d lineto ", px + pc_width, py + r1 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px, py + r3 );
	fprintf(fp, "%d %d lineto ", px + pc_width, py + r3 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c1, py );
	fprintf(fp, "%d %d lineto ", px + c1, py + r1 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c1, py + r3 );
	fprintf(fp, "%d %d lineto ", px + c1, py + pc_height );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c2, py );
	fprintf(fp, "%d %d lineto ", px + c2, py + r1 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c2, py + r3 );
	fprintf(fp, "%d %d lineto ", px + c2, py + pc_height );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r2 + pc_textup );
	fprintf(fp, "(START) show\n");

	fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r3 + pc_textup );
	fprintf(fp, "(%s) show\n", days[0].s);
    }
    
    for ( x = minX ; x <= maxX ; x++)
    {
	px = ax * x + bx;
	for ( cell = network_cols[x] ; cell != NULL ; cell = cell->next )
	{
	    if ( cell->isSpace )
		continue;
	    
	    py = ay * cell->task->ny + by;

	    fprintf(fp, "%d %d %d %d BACKGROUND\n",
		    px, px + pc_width, py, py + pc_height);

	    fprintf(fp, "%d %d %d %d OUTLINE\n",
		    px, px + pc_width, py, py + pc_height);

	    fprintf(fp, "%d %d moveto ", px, py + r1 );
	    fprintf(fp, "%d %d lineto ", px + pc_width, py + r1 );
	    fprintf(fp, "stroke\n");
	    
	    fprintf(fp, "%d %d moveto ", px, py + r3 );
	    fprintf(fp, "%d %d lineto ", px + pc_width, py + r3 );
	    fprintf(fp, "stroke\n");
	    
	    fprintf(fp, "%d %d moveto ", px + c1, py );
	    fprintf(fp, "%d %d lineto ", px + c1, py + r1 );
	    fprintf(fp, "stroke\n");
	    
	    fprintf(fp, "%d %d moveto ", px + c1, py + r3 );
	    fprintf(fp, "%d %d lineto ", px + c1, py + pc_height );
	    fprintf(fp, "stroke\n");
	    
	    fprintf(fp, "%d %d moveto ", px + c2, py );
	    fprintf(fp, "%d %d lineto ", px + c2, py + r1 );
	    fprintf(fp, "stroke\n");
	    
	    fprintf(fp, "%d %d moveto ", px + c2, py + r3 );
	    fprintf(fp, "%d %d lineto ", px + c2, py + pc_height );
	    fprintf(fp, "stroke\n");
	    
	    fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r2 + pc_textup );
	    if (task_ids)
		sprintf(buf, "%s %s", cell->task->id, cell->task->name);
	    else
		sprintf(buf, "%s", cell->task->name);
	    fprintf(fp, "(%s) show\n", buf);

	    fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r1 + pc_textup );
	    fprintf(fp, "(%s) show\n", cell->task->assigned->id);

	    fprintf(fp, "%d %d moveto ", px + c1 + pc_textin,
		    py + r3 + pc_textup );
	    sprintf(buf, "%10d", cell->task->fullduration);
	    fprintf(fp, "(%s) show\n", buf);
	    
	    fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r3 + pc_textup );
	    fprintf(fp, "(%s) show\n", days[cell->task->start].s);
	    
	    fprintf(fp, "%d %d moveto ", px + c2 + pc_textin, py + r3 + pc_textup );
	    fprintf(fp, "(%s) show\n", days[cell->task->finish].s);

	    fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r0 + pc_textup );
	    fprintf(fp, "(%s) show\n", days[cell->task->lstart].s);
	    
	    fprintf(fp, "%d %d moveto ", px + c1 + pc_textin,
		    py + r0 + pc_textup );
	    sprintf(buf, "%10d", cell->task->slack);
	    fprintf(fp, "(%s) show\n", buf);
	    
	    fprintf(fp, "%d %d moveto ", px + c2 + pc_textin, py + r0 + pc_textup );
	    fprintf(fp, "(%s) show\n", days[cell->task->lfinish].s);
	}
    }

    /* do the finish marker */
    if ( maxX == numX - 1 )
    {
	x = numX - 1;
	px = ax *  x + bx;
	py = ay * finishTL->task->ny + by;
	fprintf(fp, "%d %d %d %d BACKGROUND\n",
		px, px + pc_width, py, py + pc_height);

	fprintf(fp, "%d %d %d %d OUTLINE\n",
		px, px + pc_width, py, py + pc_height);

	fprintf(fp, "%d %d moveto ", px, py + r1 );
	fprintf(fp, "%d %d lineto ", px + pc_width, py + r1 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px, py + r3 );
	fprintf(fp, "%d %d lineto ", px + pc_width, py + r3 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c1, py );
	fprintf(fp, "%d %d lineto ", px + c1, py + r1 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c1, py + r3 );
	fprintf(fp, "%d %d lineto ", px + c1, py + pc_height );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c2, py );
	fprintf(fp, "%d %d lineto ", px + c2, py + r1 );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c2, py + r3 );
	fprintf(fp, "%d %d lineto ", px + c2, py + pc_height );
	fprintf(fp, "stroke\n");
	    
	fprintf(fp, "%d %d moveto ", px + c0 + pc_textin, py + r2 + pc_textup );
	fprintf(fp, "(FINISH) show\n");

	fprintf(fp, "%d %d moveto ", px + c2 + pc_textin, py + r3 + pc_textup );
	fprintf(fp, "(%s) show\n", days[finishDay].s);
    }
    
    /* do the milestones */
    write_milestones(fp, ax, bx, ay, by);

    fprintf(fp, "showpage\n");
    fclose(fp);
}


void NetworkDiagram(int start, int finish, char *filename)
{
    TASKLIST *tl;
    TASK *task;
//    int x;
//    NETWORK_CELL *cell;

    // printf("NETWORK chart into '%s'\n", filename);

    /* go through assigning x values as max of children + 1 */
    do
    {
	task = NULL;
	for ( tl = tasklist ; tl != NULL ; tl = tl->next )
	    if ( canAssignX(tl->task) )
	    {
		task = tl->task;
		break;
	    }

	if ( task != NULL )
	    assignX(task);
    }
    while ( task != NULL );


    /* now check that they are all done */
    for ( tl = tasklist ; tl != NULL ; tl = tl->next )
    {
	if ( strncasecmp(tl->task->id,"vac_",4) == 0 )
	    continue;
	if ( tl->task->nx == -1 )
	    Error("NetworkDiagram: task %s still has nx == -1", tl->task->id);
    }

    /* assign y values */
    assignY();

    /* construct start and finish markers */
    constructStartFinish();

    /* do the chart */
    write_chart(start, finish, filename);

//    for ( x = 0 ; x < numX ; x++ )
//	for ( cell = network_cols[x] ; cell != NULL ; cell = cell->next )
//	    printf("%s -> %d,%d\n", cell->task->id, x, cell->task->ny);
}

