#ifndef _SCHED_H
#define _SCHED_H

#include "base.h"

struct reslist;
struct task;

enum timetype
{
    MANUAL_SCHEDULE,
    AUTO_SCHEDULE,
    WORK_DONE,
    INVALID
};

typedef struct timeblock
{
    enum timetype type;
    int start;
    int finish;
    struct task *t;
    struct timeblock *next;
}
TIMEBLOCK;

typedef struct note
{
    char *text;
    struct note *next;
}
NOTE;

typedef struct resource
{
    char *id;                   // unique ID
    char *name;                 // full name
    double efficiency;          // scale to apply to task durations
    int is_group;               // is it a group?
    struct reslist *belongs_to; // if it belongs to groups - the parents
    struct reslist *contains;   // if it's a group of resources - its children
    TIMEBLOCK *booked;          // when is the resource available?
    TIMEBLOCK *rbooked;         // when is the resource available (reverse)?
    NOTE *notes;                // notes about this resource
    double rate;                // dollars per day for this resource
}
RESOURCE;


typedef struct item
{
    double cost;
    char *description;

    struct item *next;
}
ITEM;

typedef struct reslist
{
    RESOURCE *res;
    struct reslist *next;
}
RESLIST;

typedef struct tasklist
{
    struct task *task;
    struct tasklist *next;
}
TASKLIST;

typedef struct task
{
    char *id;          // unique ID
    char *name;        // printed name
    char *desc;        // longer name
    char *overrun;     // text that describes why the task is running late
    int  duration;     // number of days of work
    int  fullduration; // adjusted for efficiency of the resource and slippage
    int  origfullduration; // fullduration before slippage adjustment
    int  remaining;    // used during schedule
    int  rremaining;   // used during schedule
    int  start;        // when the assigned resource starts doing this
    int  finish;       // the last day of work
    int  lstart;       // last day assigned resource can start doing this
    int  lfinish;      // last day assigned resource can finish this
    int  slack;        // float in dates
    int  astart;       // Actual Start
    int  afinish;      // Actual Finish
    int  bstart;       // Baseline Start
    int  bfinish;      // Baseline Finish
    int  children;     // number of dependencies -- increases scheduling priority
    int  x1,y1,x2,y2;  // gannt chart drawing coordinates
    int  Nresources;   // number of people who can do this.
    int  block;        // if 1, must be scheduled as a consecutive block of time
    int  scheduled;    // if 1, has been scheduled
    int  rscheduled;   // if 1, has been scheduled on reverse run
    double percent_complete;  // fraction of this task already finished
    RESOURCE *assigned;// who is doing this?
    RESLIST *cando;    // who *might* be assigned to do this?
    TASKLIST *depends; // what other tasks must be finished first?
    TIMEBLOCK *when;   // when is this task scheduled to be done?
    TIMEBLOCK *rwhen;  // when is this task scheduled to be done (reverse)?
    NOTE *notes;       // notes about this task.
    int nx, ny;        // network diagram drawing coordinates
    int nDays;         // number of days worked so far
    TASKLIST *follows; // what other tasks will be finished after?
    int parents;       // number of follows
    ITEM *items;       // list of materials associated with task
}
TASK;

typedef struct milestone
{
    char *id;          // unique ID
    char *name;        // printed name
    int  day;          // Baseline date met
    int  bday;         // Baseline date met
    int  x1,y1,x2,y2;  // gannt chart drawing coordinates
    TASKLIST *depends; // what tasks must be finished first?
    TASK *critical;    // which depends task sets date
    int nx, ny;        // network diagram drawing coordinates
}
MILESTONE;

typedef struct milestonelist
{
    struct milestone *milestone;
    struct milestonelist *next;
}
MILESTONELIST;

typedef struct dayname
{
    time_t t;
    char *s;
    int  num;
}
DAYNAME;

typedef struct taskgraph
{
    int start;
    int finish;
    char *filename;
    struct taskgraph *next;
}
TASKGRAPH;

typedef struct tasknet
{
    int start;
    int finish;
    char *filename;
    struct tasknet *next;
}
TASKNET;

typedef enum reporttype
{
    TXT_REPORT,
    TEX_REPORT,
    HTML_REPORT
}
REPORTTYPE;

void LoadFile(char *name);
int FileIsIllegal();
FILE *OpenOutputFile(char *filename);
void PrintFile();

void DoScheduling();

void LoadDays();
void TextReport(char *fname);
void TexReport(char *filename);
void HTMLReport(char *filename);
void WeeklyReport(char *filename, REPORTTYPE rep_type);
void MonthlyReport(char *filename, REPORTTYPE rep_type);
void HardSchedule(char *filename);
void UtilGraph(char *filename);
void TaskGraph(int start, int finish, char *filename);
void NetworkDiagram(int start, int finish, char *filename);
void CostReport(char *filename, REPORTTYPE rep_type);

int FindDay(char *name);
void SortTasks();
int Mday(time_t t);
int Month(time_t t);
int Wday(time_t t);
void AddTimeBlock( TASK *task, TIMEBLOCK **t, int tstart, int tfinish,
                   enum timetype type );
void BookResource( TASK *t, RESOURCE *parent, RESOURCE *r,
                   int tstart, int tfinish, enum timetype type );
void ReverseBookResource( TASK *t, RESOURCE *parent, RESOURCE *r,
			  int tstart, int tfinish, enum timetype type );
void AssignResource(TASK *t, RESOURCE *r, int tstart);
int do_match(char *str, char *regexp, int case_sig);

int ResourceIsAvailable( RESOURCE *r, int start, int finish );
int ResourceIsUsed( RESOURCE *r, int dayNo );
int ResourceIsUsedForTask( RESOURCE *r, TASK *t, int dayNo );
int Overlap(int x1, int x2, int y1, int y2);
int DayBooked( TIMEBLOCK *tb, int dayNo );
char *TeXFix(char *buf, char *txt);
char *HTMLFix(char *buf, char *txt);
RESOURCE *FindResource(char *id);
void SlippageReport(void);

#endif

